unit PropUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, SCControl, SCStatusBar, SCDrawingSurface, SCDrawingEditor,
  SCDrawingShapes, SCDrawingCommons, SCPanels, StdCtrls, SCPopupColors,
  SCEdits, SCMaskEdit, SCAdvEdits, SCFontsAndColors, SCSpinEdits,
  SCGraphicButton, ExtDlgs, JPeg, ComCtrls, ImgList, ToolWin;

type
  TPropForm = class(TForm)
    Panel1: TPanel;
    SCControlSizer1: TSCControlSizer;
    ColorDialog: TColorDialog;
    OpenPictureDialog1: TOpenPictureDialog;
    SCGroupContainer2: TSCGroupContainer;
    SCAdvPanel4: TSCAdvPanel;
    SCAdvPanel5: TSCAdvPanel;
    TreeView1: TTreeView;
    Panel2: TPanel;
    SCGroupContainer1: TSCGroupContainer;
    SCAdvPanel1: TSCAdvPanel;
    Label7: TLabel;
    Label8: TLabel;
    Label9: TLabel;
    SCSpeedButton1: TSCSpeedButton;
    SCFrameEdit1: TSCFrameEdit;
    SCFrameEdit2: TSCFrameEdit;
    SCFrameEdit3: TSCFrameEdit;
    SCAdvPanel2: TSCAdvPanel;
    Label1: TLabel;
    Label2: TLabel;
    SCBrushStyleCombobox1: TSCBrushStyleCombobox;
    SCPopupColors1: TSCPopupColors;
    SCAdvPanel3: TSCAdvPanel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    SCPopupColors2: TSCPopupColors;
    SCPenStyleCombobox1: TSCPenStyleCombobox;
    SCPenModeCombobox1: TSCPenModeCombobox;
    SCIntSpinEdit1: TSCIntSpinEdit;
    ToolBar1: TToolBar;
    ImageList1: TImageList;
    ToolButton1: TToolButton;
    ToolButton2: TToolButton;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure SCPopupColors1MoreButtonClick(Sender: TObject);
    procedure SCPopupColors1SelectedColorChange(Sender: TObject);
    procedure SCPopupColors2MoreButtonClick(Sender: TObject);
    procedure SCPopupColors2SelectedColorChange(Sender: TObject);
    procedure SCBrushStyleCombobox1SetStyle(Sender: TObject);
    procedure SCPenStyleCombobox1SetStyle(Sender: TObject);
    procedure SCPenModeCombobox1SetMode(Sender: TObject);
    procedure SCIntSpinEdit1Change(Sender: TObject);
    procedure SCSpeedButton1Click(Sender: TObject);
    procedure SCFrameEdit1Exit(Sender: TObject);
    procedure SCFrameEdit2Exit(Sender: TObject);
    procedure SCFrameEdit3Exit(Sender: TObject);
    procedure SCFrameEdit1KeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure SCFrameEdit2KeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure SCFrameEdit3KeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormDeactivate(Sender: TObject);
    procedure TreeView1Edited(Sender: TObject; Node: TTreeNode;
      var S: String);
    procedure ToolButton2Click(Sender: TObject);
    procedure ToolButton1Click(Sender: TObject);
  private
    FUpdating: Integer;
    FCAD: TSCDeEditor;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure SetCAD(CAD: TSCDeEditor);
    procedure UpdateProps;
    procedure UpdateLayers;
  end;

var
  PropForm: TPropForm;

implementation

type
  TSCFakeShape = class(TSCDeShapeBase);

{$R *.DFM}

constructor TPropForm.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  PropForm := Self;
end;

destructor TPropForm.Destroy;
begin
  PropForm := nil;
  inherited Destroy;
end;

procedure TPropForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Action := caFree;
end;

procedure TPropForm.UpdateProps;
var
  Pm: TPenMode;
  I, Pw: Integer;
  Ps: TPenStyle;
  Bs: TBrushStyle;
  S: TSCFakeShape;
  Bc, Pc: TColor;
  N, T, C: String;
  PenFound, BrushFound,
  NameFound, CaptionFound, TextFound: Boolean;
begin
  Inc(FUpdating);
  try
    Bs := FCAD.DefaultBrush.Style;
    Bc := FCAD.DefaultBrush.Color;

    Ps := FCAD.DefaultPen.Style;
    Pm := FCAD.DefaultPen.Mode;
    Pc := FCAD.DefaultPen.Color;
    Pw := FCAD.DefaultPen.Width;

    T := '';
    N := '';
    C := '';
    
    if FCAD.SelectionCount > 0 then
    begin
      PenFound := False;
      BrushFound := False;
      CaptionFound := False;
      NameFound := False;
      TextFound := False;

      for I := 0 to FCAD.SelectionCount-1 do
      begin
        S := TSCFakeShape(FCAD.Selections[I]);

        if scssUsesBrush in S.ShapeStyle then
        begin
          if not BrushFound then
          begin
            BrushFound := True;

            Bs := S.Brush.Style;
            Bc := S.Brush.Color;
          end else
          begin
            if Bs <> S.Brush.Style then Bs := bsSolid;
            if Bc <> S.Brush.Color then Bc := clNone;
          end;
        end;

        if scssUsesPen in S.ShapeStyle then
        begin
          if not PenFound then
          begin
            PenFound := True;

            Ps := S.Pen.Style;
            Pm := S.Pen.Mode;
            Pc := S.Pen.Color;
            Pw := S.Pen.Width;
          end else
          begin
            if Ps <> S.Pen.Style then Ps := psSolid;
            if Pm <> S.Pen.Mode then Pm := pmCopy;
            if Pc <> S.Pen.Color then Pc := clNone;
            if Pw <> S.Pen.Width then Pw := 0;
          end;
        end;

        if scssUsesCaption in S.ShapeStyle then
        begin
          if not CaptionFound then
          begin
            CaptionFound := True;
            C := S.Caption;
          end else
          if C <> S.Caption then
            C := '';
        end;

        if not TextFound then
        begin
          TextFound := True;
          T := S.Text;
        end else
        if T <> S.Text then
          T := '';

        if not NameFound then
        begin
          NameFound := True;
          N := S.Name;
        end else
        if N <> S.Name then
          N := '';
      end;
    end;

    SCFrameEdit1.Text := N;
    SCFrameEdit2.Text := T;
    SCFrameEdit3.Text := C;

    SCBrushStyleCombobox1.SelectedStyle := Bs;
    SCPopupColors1.SelectedColor := Bc;

    SCPenStyleCombobox1.SelectedStyle := Ps;
    SCPenModeCombobox1.SelectedMode := Pm;
    SCPopupColors2.SelectedColor := Pc;
    SCIntSpinEdit1.IntValue := Pw;
  finally
    Dec(FUpdating);
  end;
end;

procedure TPropForm.SCPopupColors1MoreButtonClick(Sender: TObject);
begin
  ColorDialog.Color := SCPopupColors1.SelectedColor;
  if ColorDialog.Execute then
    SCPopupColors1.SelectedColor := ColorDialog.Color;
end;

procedure TPropForm.SCPopupColors1SelectedColorChange(Sender: TObject);
var
  Bc: TColor;
  I: Integer;
  S: TSCFakeShape;
begin
  Bc := SCPopupColors1.SelectedColor;
  SCBrushStyleCombobox1.BrushColor := Bc;

  if FUpdating = 0 then
  begin
    if FCAD.SelectionCount = 0 then
      FCAD.DefaultBrush.Color := Bc
    else begin
      FCAD.BeginUpdate;
      try
        for I := FCAD.SelectionCount-1 downto 0 do
        begin
          S := TSCFakeShape(FCAD.Selections[I]);
          if scssUsesBrush in S.ShapeStyle then
            S.Brush.Color := Bc;
        end;
      finally
        FCAD.EndUpdate;
      end;
    end;
  end;
end;

procedure TPropForm.SCPopupColors2MoreButtonClick(Sender: TObject);
begin
  ColorDialog.Color := SCPopupColors2.SelectedColor;
  if ColorDialog.Execute then
    SCPopupColors2.SelectedColor := ColorDialog.Color;
end;

procedure TPropForm.SCPopupColors2SelectedColorChange(Sender: TObject);
var
  Pc: TColor;
  I: Integer;
  S: TSCFakeShape;
begin
  Pc := SCPopupColors2.SelectedColor;
  SCPenStyleCombobox1.PenColor := Pc;

  if FUpdating = 0 then
  begin
    if FCAD.SelectionCount = 0 then
      FCAD.DefaultPen.Color := Pc
    else begin
      FCAD.BeginUpdate;
      try
        for I := FCAD.SelectionCount-1 downto 0 do
        begin
          S := TSCFakeShape(FCAD.Selections[I]);
          if scssUsesPen in S.ShapeStyle then
            S.Pen.Color := Pc;
        end;
      finally
        FCAD.EndUpdate;
      end;
    end;
  end;
end;

procedure TPropForm.SetCAD(CAD: TSCDeEditor);
begin
  FCAD := CAD;
end;

procedure TPropForm.SCBrushStyleCombobox1SetStyle(Sender: TObject);
var
  I: Integer;
  S: TSCFakeShape;
  Bs: TBrushStyle;
begin
  if FUpdating = 0 then
  begin
    Bs := SCBrushStyleCombobox1.SelectedStyle;

    if FCAD.SelectionCount = 0 then
      FCAD.DefaultBrush.Style := Bs
    else begin
      FCAD.BeginUpdate;
      try
        for I := FCAD.SelectionCount-1 downto 0 do
        begin
          S := TSCFakeShape(FCAD.Selections[I]);
          if scssUsesBrush in S.ShapeStyle then
            S.Brush.Style := Bs;
        end;
      finally
        FCAD.EndUpdate;
      end;
    end;
  end;
end;

procedure TPropForm.SCPenStyleCombobox1SetStyle(Sender: TObject);
var
  I: Integer;
  Ps: TPenStyle;
  S: TSCFakeShape;
begin
  if FUpdating = 0 then
  begin
    Ps := SCPenStyleCombobox1.SelectedStyle;

    if FCAD.SelectionCount = 0 then
      FCAD.DefaultPen.Style := Ps
    else begin
      FCAD.BeginUpdate;
      try
        for I := FCAD.SelectionCount-1 downto 0 do
        begin
          S := TSCFakeShape(FCAD.Selections[I]);
          if scssUsesPen in S.ShapeStyle then
            S.Pen.Style := Ps;
        end;
      finally
        FCAD.EndUpdate;
      end;
    end;
  end;
end;

procedure TPropForm.SCPenModeCombobox1SetMode(Sender: TObject);
var
  I: Integer;
  Pm: TPenMode;
  S: TSCFakeShape;
begin
  if FUpdating = 0 then
  begin
    Pm := SCPenModeCombobox1.SelectedMode;

    if FCAD.SelectionCount = 0 then
      FCAD.DefaultPen.Mode := Pm
    else begin
      FCAD.BeginUpdate;
      try
        for I := FCAD.SelectionCount-1 downto 0 do
        begin
          S := TSCFakeShape(FCAD.Selections[I]);
          if scssUsesPen in S.ShapeStyle then
            S.Pen.Mode := Pm;
        end;
      finally
        FCAD.EndUpdate;
      end;
    end;
  end;
end;

procedure TPropForm.SCIntSpinEdit1Change(Sender: TObject);
var
  I, W: Integer;
  S: TSCFakeShape;
begin
  if FUpdating = 0 then
  begin
    W := SCIntSpinEdit1.IntValue;

    if FCAD.SelectionCount = 0 then
      FCAD.DefaultPen.Width := W
    else begin
      FCAD.BeginUpdate;
      try
        for I := FCAD.SelectionCount-1 downto 0 do
        begin
          S := TSCFakeShape(FCAD.Selections[I]);
          if scssUsesPen in S.ShapeStyle then
            S.Pen.Width := W;
        end;
      finally
        FCAD.EndUpdate;
      end;
    end;
  end;
end;

procedure TPropForm.SCSpeedButton1Click(Sender: TObject);
var
  I: Integer;
  P: TPicture;
  S: TSCFakeShape;
begin
  if (FUpdating = 0) and (FCAD.SelectionCount > 0) and OpenPictureDialog1.Execute then
  begin
    P := TPicture.Create;
    try
      P.LoadFromFile(OpenPictureDialog1.FileName);

      FCAD.BeginUpdate;
      try
        for I := FCAD.SelectionCount-1 downto 0 do
        begin
          S := TSCFakeShape(FCAD.Selections[I]);
          if scssUsesPicture in S.ShapeStyle then
            S.Picture.Assign(P);
        end;
      finally
        FCAD.EndUpdate;
      end;
    finally
      P.Free;
    end;
  end;
end;

procedure TPropForm.SCFrameEdit1Exit(Sender: TObject);
var
  I: Integer;
  S: TSCFakeShape;
begin
  if FUpdating = 0 then
  begin
    FCAD.BeginUpdate;
    try
      for I := FCAD.SelectionCount-1 downto 0 do
      begin
        S := TSCFakeShape(FCAD.Selections[I]);
        S.Name := SCFrameEdit1.Text;
      end;
    finally
      FCAD.EndUpdate;
    end;
  end;
end;

procedure TPropForm.SCFrameEdit2Exit(Sender: TObject);
var
  I: Integer;
  S: TSCFakeShape;
begin
  if FUpdating = 0 then
  begin
    FCAD.BeginUpdate;
    try
      for I := FCAD.SelectionCount-1 downto 0 do
      begin
        S := TSCFakeShape(FCAD.Selections[I]);
        S.Text := SCFrameEdit2.Text;
      end;
    finally
      FCAD.EndUpdate;
    end;
  end;
end;

procedure TPropForm.SCFrameEdit3Exit(Sender: TObject);
var
  I: Integer;
  S: TSCFakeShape;
begin
  if FUpdating = 0 then
  begin
    FCAD.BeginUpdate;
    try
      for I := FCAD.SelectionCount-1 downto 0 do
      begin
        S := TSCFakeShape(FCAD.Selections[I]);
        if scssUsesCaption in S.ShapeStyle then
          S.Caption := SCFrameEdit3.Text;
      end;
    finally
      FCAD.EndUpdate;
    end;
  end;
end;

procedure TPropForm.SCFrameEdit1KeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = VK_RETURN then
  begin
    Key := 0;
    SCFrameEdit1Exit(nil);
  end;
end;

procedure TPropForm.SCFrameEdit2KeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = VK_RETURN then
  begin
    Key := 0;
    SCFrameEdit2Exit(nil);
  end;
end;

procedure TPropForm.SCFrameEdit3KeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = VK_RETURN then
  begin
    Key := 0;
    SCFrameEdit3Exit(nil);
  end;
end;

procedure TPropForm.FormDeactivate(Sender: TObject);
begin
  if ActiveControl = SCFrameEdit1 then
    SCFrameEdit1Exit(nil)
  else
  if ActiveControl = SCFrameEdit2 then
    SCFrameEdit2Exit(nil)
  else
  if ActiveControl = SCFrameEdit3 then
    SCFrameEdit3Exit(nil);
end;

procedure TPropForm.UpdateLayers;

  procedure AddNode(Parent: TTreeNode; C: TSCDeContainer);
  var
    S: String;
    I: Integer;
    Node: TTreeNode;
  begin
    for I := 0 to C.ItemCount-1 do
    begin
      S := C.Items[I].Name;
      if S = '' then
      begin
        S := C.Items[I].ClassName + IntToStr(I);
        if S[1] = 'T' then Delete(S, 1, 1);
      end;

      Node := TreeView1.Items.AddChild(Parent, S);
      Node.Data := C.Items[I];

      if C.Items[I] is TSCDeContainer then
        AddNode(Node, TSCDeContainer(C.Items[I]));

      Parent.Expand(False);
    end;
  end;

var
  S: String;
  I: Integer;
  Node: TTreeNode;
begin
  TreeView1.Items.BeginUpdate;
  try
    TreeView1.Items.Clear;
    
    for I := 0 to FCAD.LayerCount-1 do
    begin
      S := FCAD.Layers[I].Name;
      if S = '' then S := 'Layer ' + IntToStr(I);

      Node := TreeView1.Items.Add(nil, S);
      Node.Data := FCAD.Layers[I];
      
      AddNode(Node, FCAD.Layers[I]);
    end;
  finally
    TreeView1.Items.EndUpdate;
  end;
end;

procedure TPropForm.TreeView1Edited(Sender: TObject; Node: TTreeNode;
  var S: String);
begin
  if Node.Data <> nil then
    TSCFakeShape(Node.Data).Name := S;
end;

procedure TPropForm.ToolButton2Click(Sender: TObject);
begin
  if (TreeView1.Selected <> nil) and
    (TreeView1.Selected.Data <> nil) then
    TObject(TreeView1.Selected.Data).Free;
end;

procedure TPropForm.ToolButton1Click(Sender: TObject);
begin
  FCAD.ActiveLayer := TSCDeLayer.Create(FCAD);
end;

end.
